import setuptools

setuptools.setup(
    name='HubbardMeanField',
    version='1.0',
    author='LTF.123',
    author_email='tfliu@mail.ustc.edu.cn',
    description='mean field method for hubbard model',
    long_description_content_type="""text/markdown""",
    url='http://www.issp.cas.cn/',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ]
)